# Copyright (C) 2022 PlanQK
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from dimod import SampleSet, BinaryQuadraticModel, Sampler

from .startup import Startup
from ..problem import ProblemType


class SamplingStartup(Startup[ProblemType]):
    """Population creation by sampling.

    If the starting population of a genetic algorithm is created by means of a
    sampler performing global search, it is guaranteed that the first performs
    statistically at least as good as the latter because the population quality
    must not decrease per generation.

    Args:
        desired_population_size (int): Positive, maximal number of solutions to
            find.

        sampler (dimod.Sampler): A sampler that can minimise a
            `dimod.BinaryQuadraticModel`. It is advised to perform a global
            search.

    """
    def __init__(self, desired_population_size: int, sampler: Sampler):
        assert desired_population_size > 0
        self._desired_population_size = desired_population_size
        self._sampler = sampler

    def startup(self, problem: ProblemType) -> SampleSet:
        """Creates a population by minimising the given problem.

        Args:
            problem (ProblemType): The problem that is optimised
                to create a population.

        Returns:
            A `dimod.SampleSet` that contains at most `desired_population_size`,
            distinct individuals. Note there is a high probability of small
            sizes because some solutions might be found multiple times by the
            sampler. Variable type and number of qubit coincide with `problem.bqm`.

        """
        return self._sampler.sample(
            problem,
            num_reads=self._desired_population_size
        ).aggregate() # Remove duplicates



# Copyright (C) 2022 PlanQK
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from dimod import SampleSet
from typing import Callable

from .termination import Terminator, Termination
from ..problem import ProblemType


class LowestEnergyTerminator(Terminator[ProblemType]):
    """Terminating an algorithm by checking a predicate about the current
    populations's lowest energy.

    Args:
        predicate (Callable[[float], bool]): A predicate to check about the
            lowest energy.

    """
    def __init__(self, predicate: Callable[[float], bool]):
        self._predicate = predicate

    def should_terminate(self, population: SampleSet) -> bool:
        """Checks if `predicate` holds for for the current population's lowest
        energy.

        Returns:
            `True` to terminate if the lowest energy satisfies `predicate`;
            False` to proceed if not.

        """
        return self._predicate(population.first.energy)


class LowestEnergyTermination(Termination[ProblemType]):
    """Configuration on after how many generations an algorithm terminates.

    Offers rudimentary support to terminate early due to sufficient population
    quality.

    Args:
        predicate (Callable[[float], bool]): A predicate to check about the
            lowest energy. One must be aware, that those pure values are
            unintuitive and highly problem dependent, i.e. it is impossible to
            estimate what are good values.

    """
    def __init__(self, predicate: Callable[[float], bool]):
        self._predicate = predicate

    def initialise(self, problem: ProblemType) -> Terminator[ProblemType]:
        """Configurates a `LowestEnergyTerminator[ProblemType]` with the respective predicate.

        Returns:
            A `LowestEnergyTerminator[ProblemType]` checking the respective predicate.

        """
        return LowestEnergyTerminator[ProblemType](self._predicate)



# Copyright (C) 2022 PlanQK
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np

from dimod import SampleSet, Vartype, BinaryQuadraticModel
from dimod.generators import uniform
from numpy.testing import assert_array_equal
from hypothesis.strategies import composite, integers, floats
from hypothesis.extra.numpy import arrays

from ..startup.random_startup import create_random, MAX_ITERATIONS


MAX_NUM_QUBITS = 50
MAX_POPULATION_SIZE = 100


@composite
def uniform_bqm(draw, vartype: Vartype, max_num_qubits: int = MAX_NUM_QUBITS) -> BinaryQuadraticModel:
    graph = draw(integers(0, max_num_qubits))
    return uniform(graph, vartype)


@composite
def population(draw, vartype: Vartype,
               max_population_size: int = MAX_POPULATION_SIZE,
               max_num_qubits: int = MAX_NUM_QUBITS,
               max_tries: int = MAX_ITERATIONS):

    rng = np.random
    num_qubits = draw(integers(min_value=1, max_value=max_num_qubits))
    population_size = draw(integers(min_value=1, max_value=min(max_population_size, 2**num_qubits)))

    samples = create_random(num_qubits, population_size, vartype, max_tries, rng)
    energies = draw(
        arrays(
            dtype=np.float_,
            shape=population_size,
            elements=floats(allow_infinity=False, allow_nan=False)
        ))
    occurrences = draw(
        arrays(
            dtype=np.int_,
            shape=population_size,
        ))

    return SampleSet.from_samples(samples, vartype, energies, num_occurrences=occurrences)


def assert_population_equal(population: SampleSet, expected: SampleSet):
    population_indices = np.argsort(population.record)
    expected_indices = np.argsort(expected.record)

    assert_array_equal(
        population.record.sample[population_indices],
        expected.record.sample[expected_indices]
    )
    assert_array_equal(
        population.record.energy[population_indices],
        expected.record.energy[expected_indices]
    )

def assert_no_duplicates(population: SampleSet):
    """Assert that the given population does not contain any duplicates.

    Args:
        population

    Raises:
        AssertionError: If a
    """
    expected = population.aggregate()
    assert_population_equal(population, expected)


def assert_improvement(input_population: SampleSet, output_population: SampleSet):
    """Assert that the output is at least as good as the input, i.e. there is a
    solution with less equal energy than all individuals of the input population.

    Raises:
        AssertionError:
    """
    input_energy = input_population.lowest().energy
    output_energy = output_population.lowest().energy
    assert output_energy <= input_energy, \
        f"Output energy is worse than input: {output_energy} > {input_energy}."



# Copyright (C) 2022 PlanQK
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from dataclasses import dataclass
from typing import Generic

from dimod import SampleSet
from toolz.curried import pipe

from .mutation import Mutation
from .recombination import Recombination
from .selection import Selection
from .startup import Startup
from .termination import Termination
from .problem import ProblemType


@dataclass
class GeneticAlgorithm(Generic[ProblemType]):
    """Configurable implementation of a genetic algorithm.

    Solves minimisation problems of one specific subtype of `Problem`, i.e.
    problems that can be encoded as `dimod.BinaryQuadraticModel` potentially
    with additional domain knowledge.

    Returns a set of solutions as `dimod.SampleSet`. It is thus fully compatible
    with the `dimod`-API.

    This class itself is just a template that must be filled with respective
    components to work. `mutation`, `recombination`, and `selection` must
    thereby guarantee monotonicity, i.e. after their execution their best
    individual is not worse than the one of previous population.

    Args:
        startup (Startup[problemtype])): Determines the initial population. For
            instance, it can be random or used as a warm start.

        mutation (Mutation[ProblemType]): Improves quality of the current
            population by starting a local search from its individuals.

        recombination (Recombination[ProblemType]): Procreates new individuals
            by combining existing ones.

        selection (Selection[ProblemType]): Truncates low quality solutions to
            keep the population small.

        termination (Termination[ProblemType]): Allows to stop the algorithm if
            the population has achieved a certain quality level or takes too
            long.

    Note that this class is stateless, i.e. calling `optimise` multiple times
    parallelly works without any problems, as long as its components do not
    share a common object, for instance a file.

    """
    startup: Startup[ProblemType]
    mutation: Mutation[ProblemType]
    recombination: Recombination[ProblemType]
    selection: Selection[ProblemType]
    termination: Termination[ProblemType]

    def optimise(self, problem: ProblemType) -> SampleSet:
        # Instantiate each component for a each optimisation run. Therefore, they have their own internal state that
        # should not interfere with the outside environment.
        mutator = self.mutation.initialise(problem)
        recombinator = self.recombination.initialise(problem)
        selector = self.selection.initialise(problem)
        terminator = self.termination.initialise(problem)

        # Produce a first population
        population = self.startup.startup(problem)

        while True:
            # Add new individuals by mutation and recombination.
            population = pipe(population,
                              mutator.mutate,
                              recombinator.recombine)

            # Check termination condition, e.g. existence of a good solution.
            if terminator.should_terminate(population):
                break

            # Truncate low quality individuals.
            population = selector.select(population)
        return population





# Copyright (C) 2022 PlanQK
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from dimod import SampleSet

from .selection import Selector, Selection
from ..problem import ProblemType


class IdentitySelector(Selector[ProblemType]):
    """Selection by no selection.

    """
    def select(self, population: SampleSet) -> SampleSet:
        """Passes the input population without any changes.

        """
        return population

class IdentitySelection(Selection[ProblemType]):
    """Configuration of algorithm with no selection.

    Useful for writing test cases that just need some selection.

    """
    def initialise(self, problem: ProblemType) -> Selector[ProblemType]:
        """Creates a problem independent `IdentitySelector[ProblemType]`.

        """
        return IdentitySelector()



# Copyright (C) 2022 PlanQK
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""This module contains all predefined classes to configure recombination, i.e.
how a genetic algorithm can improve its popultion quality by combining
individuals. In fact, it is implements information exchange between solution
candidates.

This module uses the abstract factory pattern to allow encapsulated state and
isolated optimisation runs. Implementations of `Recombination` create
implementations of `Recombinator`, that implement the actual functionality.
"""

from .recombination import Recombinator, Recombination
from .addition_recombination import AdditionRecombinator, AdditionRecombination
from .identity_recombination import IdentityRecombination, IdentityRecombinator
from .random_bits_recombination import RandomBitsRecombinator, RandomBitsRecombination
from .random_cluster_recombination import RandomClusterRecombinator, RandomClusterRecombination



# Copyright (C) 2022 PlanQK
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np
from dimod import SampleSet, Vartype
from numpy.random import Generator, default_rng

from .startup import Startup
from ..problem import ProblemType


MAX_ITERATIONS = 100


def create_random(num_qubits: int, size: int, vartype: Vartype, max_tries: int, rng: Generator) -> np.ndarray:
    """Creates an array of spin or bit strings randomly.

    Args:
        num_qubits (int): Positive number of qubits each string should have.

        size (int): Positive number of strings in the collection.

        vartype (dimod.Vartype): Variable type, i.e. 'SPIN' or 'BINARY'.

        max_tries (int): Number of tries the procedure performs to avoid
            duplicates until it gives up.

        rng (Generator): A random number generator.

    Returns:
        A 2D numpy array with `num_qubits` columns and at most `size` rows.
    """
    assert num_qubits > 0
    assert size > 0

    # Set of values a bit/spin can have.
    choosable = np.array(list(vartype.value), dtype='i1')

    # Create bit/strings strings randomly
    # and remove duplicates
    sample = np.unique(rng.choice(choosable, size=(size, num_qubits)), axis=0)
    tries = 1

    # If we have enough individuals return population.
    # Else generate new solutions and discard duplicates
    # After some tries give up and return incomplete population
    while len(sample) != size and tries <= max_tries:
        additional = rng.choice(choosable, size=(size-len(sample), num_qubits))
        sample = np.unique(np.concatenate([sample, additional]), axis=0)
        tries += 1

    return sample


class RandomStartup(Startup[ProblemType]):
    """Random population creation.

    A lightweight start for a genetic algorithm with a diverse population.

    Args:
        population_size (int): Positive, maximal number of individuals to create.

        rng (Generator, optional): A random number generator that can be
            specified if reproducability is needed.

    """
    def __init__(self, population_size: int, rng: Generator=default_rng()):
        assert population_size > 0
        self._population_size = population_size
        self._rng = rng

    def startup(self, problem: ProblemType) -> SampleSet:
        """Creates a random population.

        Args:
            problem (ProblemType): The problem for which a population is
                created. Only variable type and amount of qubits of its
                `BinaryQuadraticModel` matter, any additional information is
                ignored.

        Returns:
            A `dimod.SampleSet` that contains at most `population_size`,
            distinct individuals. Due to randomness the actual population
            might be smaller.

        """
        sample = create_random(
            len(problem.bqm),
            self._population_size,
            problem.bqm.vartype,
            MAX_ITERATIONS,
            self._rng
            )

        return SampleSet.from_samples_bqm(
            (sample, problem.bqm.variables),
            problem.bqm,
            aggregate_samples=True
            )



# Copyright (C) 2022 PlanQK
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


from abc import abstractmethod
from dimod.sampleset import SampleSet, concatenate

from .selection import Selection, Selector
from ..problem import ProblemType


class ParetoSelector(Selector[ProblemType]):
    """Abstract base class for selection procedures that truncate primarily
    regarding some predicate and then subordinately regarding energy.

    Preserves individuals with best energies satisfying the predicate up
    to a certain population size. If there are not enough individuals, it is
    refilled with the best non-satisfactory ones.

    Args:
        num_preserved (int): Positive, maximal number of individuals to
            preserve.

        problem (ProblemType): Problem to perform the selection for. For this
            class it likely contains domain knowledge.

    """
    def __init__(self, num_preserved: int, problem: ProblemType):
        assert num_preserved > 0
        self._num_preserved = num_preserved
        self._problem = problem

    def select(self, population: SampleSet) -> SampleSet:
        """Preserves `num_preserved` individuals with best energies satisfying
        `_sample_predicate` of the current population.

        Missing solutions are refilled with the best remaining ones. Moreover,
        the actual population size is below its target iff the input population
        is below.

        """
        # Discard individuals not satisfying predicate.
        new_population = population\
            .filter(lambda s: self._sample_predicate(self._problem, s))\
            .truncate(self._num_preserved)

        # If individuals missing, add best non satisfactory individuals
        missing = self._num_preserved - len(population)
        if missing > 0:
            addition = population\
                .filter(lambda s: not self._sample_predicate(self._problem, s))\
                .truncate(missing)
            new_population = concatenate([new_population, addition])

        return new_population


    @staticmethod
    @abstractmethod
    def _sample_predicate(problem: ProblemType, sample) -> bool:
        """Abstract static method that determines, which individuals are
        preserved, even if they do not have a good energy.

        """
        pass


class ParetoSelection(Selection[ProblemType]):
    """Abstract base class for configurating a not purely energy guided,
    truncating selection.

    The simplest selection just discards individuals with bad energies. Hence,
    only constraints that are integrated in the `dimod.BinaryQuadraticModel` are
    considered. However, there are constraints tedious to express in a
    `dimod.BinaryQuadraticModel` but easy to check for a given solution. One can
    use of this by selecting only elements that satisfy this constraint. Of
    course, one must additionally consider the energy of these to respect other
    constraints.

    This idea is implemented in this class with a possibility to filter
    individuals. Note that to enable some diversity, the best, non-satisfactory
    solutions may still appear if there a not enough satisfactory individuals.

    Args:
        num_preserved (int): Positive, maximal number of individuals in the next
            population. It can be less if the input population is too small.

            Small populations decrease time to solution, but large ones might
            have more diversity, i.e. varying individuals, which increases the
            chance of finding an optimal solution.
    """
    def __init__(self, num_preserved: int):
        assert num_preserved > 0
        self._num_preserved = num_preserved



# Copyright (C) 2022 PlanQK
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""This module contains all predefined classes to configure mutation, i.e. how a
genetic algorithm can improve the elements of its current population by an
individual local search.

This module uses the abstract factory pattern to allow encapsulated state and
isolated optimisation runs. Implementations of `Mutation` create
implementations of `Recombination`, that implement the actual functionality.

"""


from .mutation import Mutator, Mutation
from .identity_mutation import IdentityMutator, IdentityMutation
from .addition_mutation import AdditionMutator, AdditionMutation
from .sampler_mutation import SamplerMutator, SamplerMutation
from .reverse_annealing_mutation import ReverseAnnealingMutation



# Copyright (C) 2022 PlanQK
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np
from numpy.random import Generator, default_rng
from dimod import SampleSet
from typing import Callable

from .recombination import Recombinator
from .pair_recombination import PairRecombinator, PairRecombination
from ..problem import ProblemType


class RandomBitsRecombinator(PairRecombinator[ProblemType]):
    """Pairwise recombination by flipping random bits.

    Seeks locations where two, randomly mated individuals differ and randomly
    chooses and flips some of those bits

    Args:
        problem (ProblemType): The problem to perform recombination.

        num_bits (int): Positive, maximal number of bits to flip. The actual
            value is less if two individuals do not differ in that many bits.

        recombination_rate (int): Positive number suggesting the amount of mates
            each individual is assigned to.

            The actual number can be both below or above this value depending on
            the existence of an appropriate apportionment. For instance, in a
            population consisting of two elements one cannot mate with more than
            one individual.

        rng (Generator, optional): A random number generator that can be
            specified if reproducability is needed.

    """
    def __init__(self,
                 problem: ProblemType,
                 num_bits: int,
                 recombination_rate: int,
                 rng: Generator = default_rng()
                 ):

        self._problem = problem
        self._num_bits = num_bits
        super().__init__(recombination_rate, rng)

    def _mate(self, population: SampleSet, pairs: np.ndarray) -> SampleSet:
        strings = population.record.sample
        left = strings[pairs[0]]
        right = strings[pairs[1]]

        result = np.zeros((2 * len(pairs), strings.shape[1]), dtype=strings.dtype)

        for i in range(len(pairs)):
            arg_xor = np.flatnonzero(left[i] != right[i])
            num_bits = min(self._num_bits, len(arg_xor))
            bits = self._rng.choice(arg_xor, size=num_bits)

            result[i] = left[i]
            result[i][bits] = right[i][bits]

            j = 2 * i
            result[j] = right[i]
            result[j][bits] = left[i][bits]

        return SampleSet.from_samples_bqm(
            (result, self._problem.bqm.variables),
            self._problem.bqm,
            aggregate_samples=True
            )


class RandomBitsRecombination(PairRecombination[ProblemType]):
    """Configuration of pairwise recombination by flipping random bits.

    Differing bits are randomly choosen and flipped for random pairs of
    individuals. It is a simple, uninformed recombination procedure.

    Does not guarantee monotonicity.

    Args:
        num_bits (int): Positive, maximal number of bits to flip. The actual
            value is less if two individuals do not differ in that many bits.

        recombination_rate (int): Positive number suggesting the amount of mates
            each individual is assigned to.

            The actual number can be both below or above this value depending on
            the existence of an appropriate apportionment. For instance, in a
            population consisting of two elements one cannot mate with more than
            one individual.

        rng (Callable[[], Generator]], optional): A function that creates an
            isolated random number generator that can be specified if
            reproducability is needed.

    """
    def __init__(self, num_bits: int, recombination_rate: int, rng: Callable[[], Generator]):
        self._num_bits = num_bits
        super().__init__(recombination_rate, rng)

    def initialise(self, problem: ProblemType) -> Recombinator[ProblemType]:
        """Configures a `RandomBitsRecombination[ProblemType]` with respective
        `num_bits` and `recombination_rate`.

        """
        return RandomBitsRecombinator[ProblemType](
            problem,
            self._num_bits,
            self._recombination_rate,
            self._rng()
            )



# Copyright (C) 2022 PlanQK
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np
from dimod import SampleSet, BinaryQuadraticModel
from numpy.random import Generator, default_rng

from .recombination import Recombinator
from .pair_recombination import PairRecombinator, PairRecombination
from ..problem import ProblemType


class RandomClusterRecombinator(PairRecombinator[ProblemType]):
    """Pairwise recombination by flipping random regons bits.

    Seeks coherent regions where two, randomly mated individuals differ,
    randomly chooses such a region and flips all of its bits.

    Args:
        problem (ProblemType): The problem to perform recombination.

        recombination_rate (int): Positive number suggesting the amount of mates
            each individual is assigned to.

            The actual number can be both below or above this value depending on
            the existence of an appropriate apportionment. For instance, in a
            population consisting of two elements one cannot mate with more than
            one individual.

        rng (Generator, optional): A random number generator that can be
            specified if reproducability is needed.

    """
    def __init__(self, problem: ProblemType, recombination_rate: int, rng: Generator = default_rng()):
        self._problem = problem
        super().__init__(recombination_rate, rng)

    def _mate(self, population: SampleSet, pairs: np.ndarray) -> SampleSet:
        strings = population.record.sample
        left = strings[pairs[0]]
        right = strings[pairs[1]]

        result = np.zeros((2 * len(pairs), strings.shape[1]), dtype=strings.dtype)

        for i in range(len(pairs)):
            arg_xor = np.flatnonzero(left[i] != right[i])
            bit = self._rng.choice(arg_xor, size=1)[0]

            j = 2 * i
            result[i] = left[i]
            result[j] = right[i]

            cluster_start = bit
            while cluster_start >= 0 and right[i][cluster_start] != left[i][cluster_start]:
                result[i][cluster_start] = right[i][cluster_start]
                result[j][cluster_start] = left[i][cluster_start]
                cluster_start -= 1

            cluster_end = bit
            while cluster_end < strings.shape[1] and right[i][cluster_end] != left[i][cluster_end]:
                result[i][cluster_end] = right[i][cluster_end]
                result[j][cluster_end] = left[i][cluster_end]
                cluster_end += 1


        return SampleSet.from_samples_bqm(
            (result, self._problem.bqm.variables),
            self._problem.bqm,
            aggregate_samples=True
            )

class RandomClusterRecombination(PairRecombination[ProblemType]):
    """Configuration of pairwise recombination by flipping random regions bits.

    Coherent regions of differing bits are randomly choosen and flipped for
    random pairs of individuals. It is a simple, uninformed recombination
    procedure.

    Does not guarantee monotonicity.

    Args:
        recombination_rate (int): Positive number suggesting the amount of mates
            each individual is assigned to.

            The actual number can be both below or above this value depending on
            the existence of an appropriate apportionment. For instance, in a
            population consisting of two elements one cannot mate with more than
            one individual.

        rng (Callable[[], Generator]], optional): A function that creates an
            isolated random number generator that can be specified if
            reproducability is needed.

    """
    def initialise(self, problem: ProblemType) -> Recombinator[ProblemType]:
        """Configures a `RandomClusterRecombination[ProblemType]` with
        respective recombination_rate`.

        """
        return RandomClusterRecombinator[ProblemType](
            problem,
            self._recombination_rate,
            self._rng()
            )



# Copyright (C) 2022 PlanQK
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from dimod import SampleSet
from typing import List

from .termination import Terminator, Termination
from ..problem import ProblemType


class OrTerminator(Terminator[ProblemType]):
    """Terminating an algorithm if any of multiple conditions holds.

    Args:
        children (List[Terminator[ProblemType]])

    """
    def __init__(self, children: List[Terminator[ProblemType]]):
        self._children = children

    def should_terminate(self, population: SampleSet) -> bool:
        """Checks if any child requests termination for the current
        population.

        Returns:
            `True` to terminate if any child returns `True`; `False` to
            proceed if all children return `False`.

        """
        return any(s.should_terminate(population) for s in self._children)


class OrTermination(Termination[ProblemType]):
    """Configuration of a selection of termination conditions.

    Only one termination condition must hold to halt.

    Args:
        children (List[Termination[ProblemType]])

    """
    def __init__(self, children: List[Termination[ProblemType]]):
        self._children = children

    def initialise(self, problem: ProblemType) -> Terminator[ProblemType]:
        """Problem-specific configuration of termination.

        Args:
            problem (ProblemType): A concrete problem instance,
                that is used to configure the children.

        Returns:
            An `OrTerminator[ProblemType]` that checks if any child condition holds.

        """
        return OrTerminator[ProblemType]([s.initialise(problem) for s in self._children])



# Copyright (C) 2022 PlanQK
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from abc import ABC, abstractmethod
from typing import Generic

from dimod import SampleSet

from ..problem import ProblemType


class Startup(Generic[ProblemType], ABC):
    """Interface for creating the initial population.
    """
    @abstractmethod
    def startup(self, problem: ProblemType) -> SampleSet:
        """Creates a population.

        Args:
            problem (ProblemType): The problem for which a
                population is created.

        Returns:
            A non-empty `dimod.SampleSet` that shares the same variable type and
            amount of qubits with `problem`.

        """



# Copyright (C) 2022 PlanQK
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from abc import ABC, abstractmethod
from typing import Generic

from dimod import SampleSet
from ..problem import ProblemType


class Recombinator(Generic[ProblemType], ABC):
    """An interface for combining individuals.

    """
    @abstractmethod
    def recombine(self, population: SampleSet) -> SampleSet:
        """Combines individuals to create a new population.

        Args: population (dimod.SampleSet): Nonempty population from which
            individuals are chosen to recombine.

        Returns:
            A nonempty `dimod.SampleSet`.

        """
        pass


class Recombination(Generic[ProblemType], ABC):
    """Interface to configure recombination procedure of a genetic algorithm.

    It determines how individuals are combined to create new improved ones.

    """
    @abstractmethod
    def initialise(self, problem: ProblemType) -> Recombinator[ProblemType]:
        """Problem-specific configuration of recombination.

        Args:
            problem (ProblemType): A concrete problem instance,
                for which a bespoke recombination can be prepared.

        Returns:
            A `Recombinator[ProblemType]` implementation.
        """
        pass



# Copyright (C) 2022 PlanQK
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np
from abc import abstractmethod
from numpy.random import Generator, default_rng
from typing import Callable

from dimod import SampleSet

from .recombination import Recombinator, Recombination
from ..problem import ProblemType


def create_pairs(population_size: int, recombination_rate: int, rng: Generator = default_rng()):
    """Recombination of pairs.

    Tries to choose randomly `recombination_rate` mates for each `population_size` elements.
    An element cannot be mated with itself and there are no dublicate pairs.

    Args:
        population_size: Number of individuals that should be mated. There must be at least two.
        recombination_rate: Amount of mates each individual should be with paired.
        rng: Random number generator.
    """
    assert population_size > 1
    assert recombination_rate >= 1

    # If recombination rate is too large, create all possible pairs
    if population_size <= recombination_rate:
        recombination_rate = population_size - 1


    # Stores pairs
    assignment = np.zeros((2, population_size * recombination_rate), dtype=int)
    # Stores which mates are choosable in the respective iteration
    assignable = np.arange(1, population_size, dtype=int)

    # Sample mates for first individual
    assignment[1, :recombination_rate] = rng.choice(assignable, size=recombination_rate, replace=False)

    chosen = recombination_rate
    for individual in range(1, population_size):
        # Prohibit choosing one self
        assignable[individual-1] -= 1

        # Sample mates for current individual
        new_chosen = chosen + recombination_rate
        assignment[0, chosen:new_chosen] = individual
        assignment[1, chosen:new_chosen] = rng.choice(assignable, size=recombination_rate, replace=False)

        chosen = new_chosen

    # Eliminate duplicates
    assignment = np.unique(np.sort(assignment, axis=0), axis=1)

    return assignment


class PairRecombinator(Recombinator[ProblemType]):
    """Abstract class for combining pairs of individuals.

    It implements random creation of pairs, that subclasses can use to recombine
    new individuals from them.

    Args:
        recombination_rate (int): Positive number suggesting the amount of mates
            each individual is assigned to.

            The actual number can be both below or above this value depending on
            the existence of an appropriate apportionment. For instance, in a
            population consisting of two elements one cannot mate with more than
            one individual.

        rng (Generator, optional): A random number generator that can be
            specified if reproducability is needed.

    """
    def __init__(self, recombination_rate: int, rng: Generator = default_rng()):
        assert recombination_rate > 0
        self._recombination_rate = recombination_rate
        self._rng = rng

    def recombine(self, population: SampleSet) -> SampleSet:
        """Combines individuals pairwise.

        First, pairs of individuals are created randomly and then `self._mate`
        is used to recombine those.

        """
        pairs = create_pairs(len(population), self._recombination_rate)
        return self._mate(population, pairs)

    @abstractmethod
    def _mate(self, population: SampleSet, pairs: np.ndarray) -> SampleSet:
        """Abstract method determining combination of pairs.

        Args:
            pairs (np.ndarray): An array of pairs indices representing the
                respective individuals. Two solutions are at most paired once
                and never with oneself.

        """
        pass


class PairRecombination(Recombination[ProblemType]):
    """Basic, abstract configuration for pairwise recombination.

    Args:
        recombination_rate (int): Positive number suggesting the amount of mates
            each individual is assigned to.

            The actual number can be both below or above this value depending on
            the existence of an appropriate apportionment. For instance, in a
            population consisting of two elements one cannot mate with more than
            one individual.

        rng (Callable[[], Generator]], optional): A function that creates an
            isolated random number generator that can be specified if
            reproducability is needed.
    """
    def __init__(self, recombination_rate: int, rng: Callable[[], np.random.Generator] = default_rng):
        assert recombination_rate > 1
        self._recombination_rate = recombination_rate
        self._rng = rng



# Copyright (C) 2022 PlanQK
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


from dimod.sampleset import SampleSet, concatenate

from .recombination import Recombinator, Recombination
from ..problem import ProblemType


class AdditionRecombinator(Recombinator[ProblemType]):
    """Adding the old population after recombination.

    An arbitrary `Recombinator[ProblemType]` is injected, that performs
    recombination. Its result is then merged with the input population.

    Args:
        internal (Recombinator[ProblemType]): Determines the recombination that
            is done first, before original individuals are added.

    """
    def __init__(self, internal: Recombinator[ProblemType]):
        self._internal = internal

    def recombine(self, population: SampleSet) -> SampleSet:
        """Performs internal recombination on the current population and merge
        it with the result.

        Returns:
            A `dimod.SampleSet` that preserves the guarantees of `internal`. In
            addition, it contains any element of `population` and no duplicates.

        """
        return concatenate([population, self._internal.recombine(population)]).aggregate()


class AdditionRecombination(Recombination[ProblemType]):
    """Configuration of a recombination reusing the old population.

    An arbitrary `Recombination[ProblemType]` is used to perform recombination,
    whose individuals are then added the input population.

    It is the easiest way to achieve monotonicity for any recombination
    procedure.

    Args:
        internal (Recombination[ProblemType]): Configures the recombination that
            is done first, before original individuals are added.

    """
    def __init__(self, internal: Recombination[ProblemType]):
        self._internal = internal

    def initialise(self, problem: ProblemType) -> Recombinator[ProblemType]:
        """Configures a `AdditionRecombination[ProblemType]` with the respective
        internal recombination.

        Args:
            problem (ProblemType): A concrete problem instance,
                that is used to configure the internal recombination.

        """
        return AdditionRecombinator[ProblemType](self._internal.initialise(problem))



# Copyright (C) 2022 PlanQK
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from dimod import SampleSet

from .selection import Selector, Selection
from ..problem import ProblemType


class TruncationSelector(Selector[ProblemType]):
    """Selection by preserving only the best individuals.

    Preserves the `num_preserved` best individuals regarding their energies and
    discards the remaining ones.

    Args:
        num_preserved (int): Positive, maximal number of individuals in the next
            population. It can be less if the input population is too small.

    """
    def __init__(self, num_preserved: int):
        assert num_preserved >= 0
        self._num_preserved = num_preserved

    def select(self, population: SampleSet) -> SampleSet:
        """Removes individuals with the worst energies.

        Args:
            population (dimod.SampleSet): Nonempty population to select from.

        Returns:
            A nonempty `dimod.SampleSet` with at most `num_preserved` elements.
            It is an at least as good subset of `population`

        """
        assert len(population) > 0
        return population.truncate(self._num_preserved)


class TruncationSelection(Selection[ProblemType]):
    """Configuration of a simple only-the-best selection.

    If an algorithm is configured with it, after each iteration, energies are
    sorted regarding their energy; the `num_preserved` best individuals are
    selected and the remaining ones are discarded.

    Monotonicity is guaranteed.

    Args:
        num_preserved (int): Positive, maximal number of individuals in the next
            population. It can be less if the input population is too small.

            Small populations decrease time to solution, but large ones might
            have more diversity, i.e. varying individuals, which increases the
            chance of finding an optimal solution.

    """
    def __init__(self, num_preserved: int):
        assert num_preserved >= 1
        self._num_preserved = num_preserved

    def initialise(self, problem: ProblemType) -> Selector[ProblemType]:
        """Configures a `TruncationSelector[ProblemType]` with the respective `num_preserved`.
        """
        return TruncationSelector[ProblemType](self._num_preserved)



# Copyright (C) 2022 PlanQK
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""This module contains all important classes and implementations of the
components that are necessary to have a quick start on how to use this library.

Note that classes in this module may differ in implementation from those in the
respective submodules with same name. This enables for the latter a more general
area of applications.

"""

from dimod import Sampler, BinaryQuadraticModel
from typing import Callable, Dict, Any
from numpy.random import Generator

from qaga import mutation as _mutation
from qaga import recombination as _recombination
from qaga import selection as _selection

from .genetic_algorithm import GeneticAlgorithm
from .startup import RandomStartup, WarmStartup, SamplingStartup
from .problem import Problem, ProblemType
from .termination import *


class SamplerMutation(_mutation.SamplerMutation[ProblemType]):
    """Configuration of individual improvement by means of local search
    annealing.

    Classically simulates an actual Quantum-Assisted Genetic Algorithm (QAGA).
    It should be the default class for testing and experimenting, as it does not
    require access to any proprietary quantum hardware.

    The results of the sampler runs are merge with the input population, thus,
    monotonicity is guaranteed.

    Args:
        sampler (dimod.Sampler): A classical sampler.

        sampler_config (dict): Possible configuration for the sampler that is
            specifically used to force local instead of global search

    """
    def __init__(self, sampler: Sampler, sampler_config: Dict[str, Any] = {}):
        mutation = _mutation.SamplerMutation(sampler, sampler_config)
        self._internal = _mutation.AdditionMutation(mutation)

    def initialise(self, problem: BinaryQuadraticModel) -> _mutation.Mutator:
        return self._internal.initialise(problem)


class ReverseAnnealingMutation(_mutation.ReverseAnnealingMutation[ProblemType]):
    """Individual improvement by means of reverse annealing.

    Lifts an ordinary genetic algorithm to a Quantum-Assisted Genetic
    Algorithm (QAGA). Reverse Annealing facilitates DWave's quantum annealer
    hardware to start a given state and perform local search. Thus it is
    predestined as a mutation procedure.

    The results of the sampler runs are merge with the input population, thus,
    monotonicity is guaranteed.

    Note:
        Usage of this class requires access to DWave's commercial samplers.
        For an alternative have a look at `qaga.mutation.SamplerMutation`,
        that offers support for free, classical samplers.

    Args:
        sampler (dimod.Sampler): One of DWave's proprietary samplers. If
            embeddings are needed, using `FixedEmbeddingComposite` or
            `LazyFixedEmbeddingComposite` is strongly encouraged to avoid
            recalculation in each generation.

        sampler_config (dict): Possible configuration for the sampler. For
            instance, to do reverse annealing an 'annealing_path' must
            be given.

    """
    def __init__(self, sampler: Sampler, sampler_config: Dict[str, Any] = {}):
        mutation = _mutation.ReverseAnnealingMutation(sampler, sampler_config)
        self._internal = _mutation.AdditionMutation(mutation)

    def initialise(self, problem: BinaryQuadraticModel) -> _mutation.Mutator[ProblemType]:
        return self._internal.initialise(problem)


class RandomBitsRecombination(_recombination.RandomBitsRecombination[ProblemType]):
    """Configuration of pairwise recombination by flipping random bits.

    Differing bits are randomly choosen and flipped for random pairs of
    individuals. It is a simple, uninformed recombination procedure.

    The results of the bit flips are merge with the input population, thus,
    monotonicity is guaranteed.

    Args:
        num_bits (int): Positive, maximal number of bits to flip. The actual
            value is less if two individuals do not differ in that many bits.

        recombination_rate (int): Positive number suggesting the amount of mates
            each individual is assigned to.

            The actual number can be both below or above this value depending on
            the existence of an appropriate apportionment. For instance, in a
            population consisting of two elements one cannot mate with more than
            one individual.

        rng (Callable[[], Generator]], optional): A function that creates an
            isolated random number generator that can be specified if
            reproducability is needed.

    """
    def __init__(self, num_bits: int, recombination_rate: int, rng: Callable[[], Generator]):
        recombination = _recombination.RandomBitsRecombination(num_bits, recombination_rate, rng)
        self._internal = _recombination.AdditionRecombination(recombination)

    def initialise(self, problem: BinaryQuadraticModel) -> _recombination.Recombinator:
        return self._internal.initialise(problem)


class RandomClusterRecombination(_recombination.RandomClusterRecombination[ProblemType]):
    """Configuration of pairwise recombination by flipping random regions bits.

    Coherent regions of differing bits are randomly choosen and flipped for
    random pairs of individuals. It is a simple, uninformed recombination
    procedure.

    The results of the bit flips are merge with the input population, thus,
    monotonicity is guaranteed.

    Args:
        recombination_rate (int): Positive number suggesting the amount of mates
            each individual is assigned to.

            The actual number can be both below or above this value depending on
            the existence of an appropriate apportionment. For instance, in a
            population consisting of two elements one cannot mate with more than
            one individual.

        rng (Callable[[], Generator]], optional): A function that creates an
            isolated random number generator that can be specified if
            reproducability is needed.

    """
    def __init__(self, recombination_rate: int, rng: Callable[[], Generator]):
        recombination = _recombination.RandomClusterRecombination(recombination_rate, rng)
        self._internal = _recombination.AdditionRecombination(recombination)

    def initialise(self, problem: BinaryQuadraticModel) -> _recombination.Recombinator:
        return self._internal.initialise(problem)


class TruncationSelection(_selection.TruncationSelection[ProblemType]):
    """Configuration of a simple only-the-best selection.

    If an algorithm is configured with it, after each iteration, energies are
    sorted regarding their energy; the `num_preserved` best individuals are
    selected and the remaining ones are discarded. However, if the population
    size is below a specific value, random individuals are added.

    This helps to prevent so called elitism, i.e. the degeneration of the
    population of later generations because all individuals have been mutated to
    the same, possibly local, minima. New, unrelated solutions widen the search
    space and increase the chance of finding global minima.

    Monotonicity is guaranteed.

    Args:

        num_preserved (int): Positive number of individuals that are preserved
            from the input population. It is guaranteed unless the solution space
            size makes this impossible. If the population size is less then
            random individuals are created.

        num_new_solutions(int): Positive number of additional, randomly created
            individuals regardless of input population size.

        rng (Callable[[], Generator]], optional): A function that creates an
            isolated random number generator that can be specified if
            reproducability is needed.

    """
    def __init__(self, num_preserved: int, num_new_solutions: int, rng: Callable[[], Generator]):
        assert num_preserved > 0
        assert num_new_solutions > 0
        recombination = _selection.TruncationSelection(num_preserved)
        self._internal = _selection.RefillingSelection(num_preserved + num_new_solutions, recombination, rng)

    def initialise(self, problem: BinaryQuadraticModel) -> _selection.Selector:
        return self._internal.initialise(problem)




# Copyright (C) 2022 PlanQK
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from abc import ABC, abstractmethod
from typing import Generic

from dimod import SampleSet
from ..problem import ProblemType


class Mutator(Generic[ProblemType], ABC):
    """An interface for improving individuals.

    """
    @abstractmethod
    def mutate(self, population: SampleSet) -> SampleSet:
        """
        Improves individuals of the calling algorithm's current population.

        Implementations of this interface must ensure that in the result population no individual appears twice.
        Moreover, the best individual of the new population must be at least as good as the best one of the old population.

        :param population: The calling algorithm's current population.
        """
        pass


class Mutation(Generic[ProblemType], ABC):
    """Interface to configure mutation procedure of a genetic algorithm.

    It determines how solution candidates are improved individually, often by
    some local search.

    """

    @abstractmethod
    def initialise(self, problem: ProblemType) -> Mutator[ProblemType]:
        """Problem-specific configuration of mutation.

        Args:
            problem (ProblemType): A concrete problem instance,
                for which a bespoke mutation can be prepared.

        Returns:
            A `Mutator[ProblemType]` implementation.
        """
        pass



# Copyright (C) 2022 PlanQK
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from dimod.sampleset import SampleSet, concatenate

from .mutation import Mutator, Mutation
from ..problem import ProblemType


class AdditionMutator(Mutator[ProblemType]):
    """Adding the old population after mutation.

    An arbitrary `Mutator[ProblemType]` is injected, that performs
    mutation. Its result is then merged with the input population.

    Args:
        internal (Mutator[ProblemType]): Determines the mutation that
            is done first, before the original individuals are added.

    """
    def __init__(self, internal: Mutator[ProblemType]):
        self._internal = internal

    def mutate(self, population: SampleSet) -> SampleSet:
        new = self._internal.mutate(population)
        return concatenate((population, new)).aggregate()


class AdditionMutation(Mutation[ProblemType]):
    """Configuration of a mutation reusing the old population.

    An arbitrary `Mutation[ProblemType]` is used to perform mutation,
    whose individuals are then added the input population.

    It is the easiest way to achieve monotonicity for any mutation
    procedure.

    Args:
        internal (Mutation[ProblemType]): Configures the mutation that
            is done first, before the original individuals are added.

    """
    def __init__(self, internal: Mutation[ProblemType]):
        self._internal = internal


    def initialise(self, problem: ProblemType) -> Mutator[ProblemType]:
        return AdditionMutator[ProblemType](self._internal.initialise(problem))



# Copyright (C) 2022 PlanQK
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from dimod import SampleSet
from copy import deepcopy

from .startup import Startup
from ..problem import ProblemType

class WarmStartup(Startup[ProblemType]):
    """Population creation by reusing an already existing one.

    If one has already a set of solutions with sufficient quality, a genetic
    algorithm can use it as initial population. This guarantees that the final
    one is at least as good as the initial one, however, there is the chance
    even better solutions are found.

    Args:
        population (dimod.SampleSet): A non-empty population to start with.

    """
    def __init__(self, population: SampleSet):
        assert len(population) > 0
        self._population = population

    def startup(self, problem: ProblemType) -> SampleSet:
        """Reuses the deposited population.

        Args:
            problem (ProblemType): The problem for which the
                deposited population is reused. Variable type and amount of
                qubits must fit.

        Returns:
            A `dimod.SampleSet` that is a deep copy of the deposited population.
            No augmentations are done.

        """
        assert self._population.variables == problem.bqm.variables
        assert self._population.vartype == problem.bqm.vartype

        return deepcopy(self._population)



# Copyright (C) 2022 PlanQK
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""This module contains all predefined classes to configure startup, i.e. how
the initial population of a genetic algorithm is created.

A good starting population can decrease runtime performance, if the individuals
are already good, and increase solution quality of it is diverse.

"""

from .startup import Startup
from .random_startup import RandomStartup
from .warm_startup import WarmStartup
from .sampling_startup import SamplingStartup



# Copyright (C) 2022 PlanQK
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from dimod import SampleSet, BinaryQuadraticModel

from .recombination import Recombinator, Recombination
from ..problem import ProblemType


class IdentityRecombinator(Recombinator[ProblemType]):
    """Recombination by no recombination.

    """
    def recombine(self, population: SampleSet) -> SampleSet:
        """Passes the input population without any changes.

        """
        return population


class IdentityRecombination(Recombination[ProblemType]):
    """Configuration of algorithm with no recombination.

    Useful for writing test cases that just need some recombination.

    """
    def initialise(self, problem: ProblemType) -> Recombinator[ProblemType]:
        """Creates a problem independent `IdentityRecombinator[ProblemType]`.

        """
        return IdentityRecombinator[ProblemType]()



# Copyright (C) 2022 PlanQK
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from dimod import Sampler
from dwave.system import ReverseBatchStatesComposite
from typing import Dict, Any

from .mutation import Mutator
from .sampler_mutation import SamplerMutator, SamplerMutation
from ..problem import ProblemType


class ReverseAnnealingMutation(SamplerMutation[ProblemType]):
    """Individual improvement by means of reverse annealing.

    Lifts an ordinary genetic algorithm to a Quantum-Assisted Genetic
    Algorithm (QAGA). Reverse Annealing facilitates DWave's quantum annealer
    hardware to start a given state and perform local search. Thus it is
    predestined as a mutation procedure.

    Does not guarantee monotonicity.

    Note:
        Usage of this class requires access to DWave's commercial samplers.
        For an alternative have a look at `qaga.mutation.SamplerMutation`,
        that offers support for free, classical samplers.

    Args:
        sampler (dimod.Sampler): One of DWave's proprietary samplers. If
            embeddings are needed, using `FixedEmbeddingComposite` or
            `LazyFixedEmbeddingComposite` is strongly encouraged to avoid
            recalculation in each generation.

        sampler_config (dict): Possible configuration for the sampler. For
            instance, to do reverse annealing an 'annealing_path' must
            be given.

    """
    def __init__(self, sampler: Sampler, sampler_config: Dict[str, Any] = {}):
        # Avoid a current bug in DWave-API
        sampler.parameters.pop("num_reads")
        super().__init__(ReverseBatchStatesComposite(sampler), sampler_config)

    def initialise(self, problem: ProblemType) -> Mutator[ProblemType]:
        """Configures and creates a new object that performs mutation
        with reverse annealing.

        Args:
            problem (ProblemType): The problem the mutation should be used for.

        Returns:
            A `SamplerMutator[ProblemType]`

        """
        return SamplerMutator[ProblemType](self._sampler, problem, self._sampler_config)



# Copyright (C) 2022 PlanQK
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from dataclasses import dataclass
from typing import TypeVar

from dimod import BinaryQuadraticModel

@dataclass
class Problem:
    """Base class for any problem to optimise.

    In general, it necessary that it contains a `dimod.BinaryQuadraticModel`
    encoding the actual problem, however, additional information can be added by
    interheritance to make use of domain knowledge.

    """
    bqm : BinaryQuadraticModel

ProblemType = TypeVar('ProblemType', bound=Problem)
"""TypeVariable of all subtypes of `Problem`.

Standardises the problem type all procedures use.

"""



# Copyright (C) 2022 PlanQK
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Callable
from dimod.sampleset import SampleSet, concatenate
from numpy.random import Generator, default_rng

from .selection import Selector, Selection
from ..startup.random_startup import create_random
from ..problem import ProblemType


MAX_FILLING_ITERATION = 1_000
""" Maximal number of iterations that are performed to generate enough distinct
individuals.

 """

class RefillingSelector(Selector[ProblemType]):
    """Guarantees a certain population size after a selection.

    An arbitrary `Selector[ProblemType]` is injected, that performs selection.
    If its result population is below a specific value, random individuals are
    added.

    Args:
        num_preserved (int): Positive number of individuals the output
            population should have. It is guaranteed unless the solution space
            size makes this impossible.

        problem (ProblemType): The concrete problem instance this
            selection is performed for.

        internal (Selector[ProblemType]): Determines the selection that is done
            first, before additional individuals are added.

        rng (Generator, optional): A random number generator that can be
            specified if reproducability is needed.

    """
    def __init__(self, num_preserved: int, problem: ProblemType,
                 internal: Selector, rng: Generator = default_rng()):
        assert num_preserved > 0
        self._num_preserved = num_preserved
        self._problem = problem
        self._internal = internal
        self._rng = rng

    def select(self, population: SampleSet) -> SampleSet:
        """Performs internal selection on the current population and refills
        with random individuals.

        If the selection of `population` does not contain at least
        `num_preserved` solutions, random ones are added till the size
        `num_preserved` is established.

        Args:
            population (dimod.SampleSet): The current population that should be
                transformed. Variable type and number of qubits must coincide
                with `problem`.

        Returns:
            A `dimod.SampleSet` that preserves the guarantees of `internal`. In
            addition, it contains `num_preserved` distinct individuals the
            solution space size makes this impossible.

        Raises:
            Exception: Due to randomness there is a chance that always already
                existing solutions are drawn, the population size stays the
                same. To guarantee termination a limited number of rerolls are
                done. If it is exceeded a exception is thrown.

        """
        # Do not try to pick more states that available.
        new_population = self._internal.select(population)

        for _ in range(MAX_FILLING_ITERATION):
            if len(new_population) == self._num_preserved:
                return new_population

            # Compute how many additional elements must be created including the replanishing ones.
            num_missing = self._num_preserved - len(new_population)

            # Create new individuals
            additional_individuals = create_random(len(self._problem.bqm), num_missing,
                                                   self._problem.bqm.vartype, MAX_FILLING_ITERATION, self._rng)
            sampleset = SampleSet.from_samples_bqm((additional_individuals, self._problem.bqm.variables),
                                                   self._problem.bqm)
            # Merge both sets removing duplicates
            new_population = concatenate([new_population, sampleset]).aggregate()

        raise Exception(
            "Tried to create new individuals "
            "but failed {MAX_FILLING_ITERATION}-times!"
        )


class RefillingSelection(Selection[ProblemType]):
    """Configure a selection to guarantee a certain population size.

    An arbitrary `Selection[ProblemType]` is used to perform selection. If its result
    population is below a specific value, random individuals are added.

    This helps to prevent so called elitism, i.e. the degeneration of the
    population of later generations because all individuals have been mutated to
    the same, possibly local, minima. New, unrelated solutions widen the search
    space and increase the chance of finding global minima.

    Monotonicity is only guaranteed if `internal` guarantees it.

    Args:
        num_preserved (int): Positive number of individuals output population
            should have. It is guaranteed unless the solution space size makes
            this impossible.

        internal (Selection[ProblemType]): Configures the selection that is done first,
            before additional individuals are added.

        rng (Callable[[], Generator]], optional): A function that creates an
            isolated random number generator that can be specified if
            reproducability is needed.

    """
    def __init__(self,
                 num_preserved: int,
                 internal: Selection,
                 rng: Callable[[], Generator] = default_rng):
        assert num_preserved > 0
        self._num_preserved = num_preserved
        self._internal = internal
        self._rng = rng

    def initialise(self, problem: ProblemType) -> Selector[ProblemType]:
        """Configures a `RefillingSelector[ProblemType]` with the respective
        `num_preserved` and internal selection.

        Args:
            problem (ProblemType): The concrete problem instance this
                selection is performed for. It determines variable type and number
                for the randomly created individuals.

        """
        return RefillingSelector(
            self._num_preserved,
            problem,
            self._internal.initialise(problem),
            self._rng()
            )



# Copyright (C) 2022 PlanQK
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from dimod import SampleSet

from .termination import Terminator, Termination
from ..problem import ProblemType


class GenerationTerminator(Terminator[ProblemType]):
    """Terminating an algorithm by bounding the number of generations.

    Args:
        num_generations: Positive number of iterations after that an algorithm
            should terminate.

    """
    def __init__(self, num_generations: int):
        assert num_generations > 0
        self._num_generations = num_generations
        self._current_generation = 0

    def should_terminate(self, population: SampleSet) -> bool:
        """Checks if itself was called at least `num_generation` times.

        Returns:
            `True` to terminate if this method was called more than
            `num_generation` times, `False` to proceed if not.

        """
        self._current_generation += 1
        return self._current_generation > self._num_generations


class GenerationTermination(Termination[ProblemType]):
    """Configuration on after how many generations an algorithm terminates.

    This is the preferred way to guarantee termination.

    Args:
        num_generations: Positive number of iterations after that an algorithm
            should terminate. If it is too small, population quality might be
            negatively affected.

    """
    def __init__(self, num_generations: int):
        assert num_generations > 0
        self.num_generations = num_generations

    def initialise(self, problem: ProblemType) -> Terminator[ProblemType]:
        """Configurates a `GenerationTerminator[ProblemType]` with the
        respective number of generations.

        """
        return GenerationTerminator[ProblemType](self.num_generations)



# Copyright (C) 2022 PlanQK
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from dimod import Sampler, SampleSet
from typing import Dict, Any

from .mutation import Mutator, Mutation
from ..problem import ProblemType


class SamplerMutator(Mutator[ProblemType]):
    """Individual improvement by means classic local search.

    Args:
        sampler (dimod.Sampler): A classical sampler.

        sampler_config (dict): Possible configuration for the sampler that is
            specifically used to force local instead of global search

    """
    def __init__(self, sampler: Sampler, problem: ProblemType, sampler_config: Dict[str, Any] = {}):
        self._sampler = sampler
        self._sampler_config = sampler_config
        self._problem = problem

    def mutate(self, population: SampleSet) -> SampleSet:
        """Performs one sampler run for each individual.

        """
        return self._sampler.sample(self._problem.bqm, initial_states=population, **self._sampler_config)


class SamplerMutation(Mutation[ProblemType]):
    """Configuration of individual improvement by means of local search
    annealing.

    Classically simulates an actual Quantum-Assisted Genetic Algorithm (QAGA).
    It should be the default class for testing and experimenting, as it does not
    require access to any proprietary quantum hardware.

    Does not guarantee monotonicity.

    Args:
        sampler (dimod.Sampler): A classical sampler.

        sampler_config (dict): Possible configuration for the sampler that is
            specifically used to force local instead of global search

    """
    def __init__(self, sampler: Sampler, sampler_config: Dict[str, Any] = {}):
        self._sampler = sampler
        self._sampler_config = sampler_config

    def initialise(self, problem: ProblemType) -> Mutator[ProblemType]:
        """Configures and creates a new object that performs mutation with an
        injected sampler.

        Args:
            problem (ProblemType): The problem the mutation should be used for.

        Returns:
            A `SamplerMutator[ProblemType]`

        """
        return SamplerMutator[ProblemType](self._sampler, problem, self._sampler_config)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['qaga',
 'qaga.mutation',
 'qaga.recombination',
 'qaga.selection',
 'qaga.startup',
 'qaga.termination',
 'qaga.testing']

package_data = \
{'': ['*']}

install_requires = \
['dimod>=0.10.15,<0.11.0',
 'dwave-neal>=0.5.9,<0.6.0',
 'dwave-system>=1.11.0,<2.0.0',
 'toolz>=0.11.2,<0.12.0']

setup_kwargs = {
    'name': 'qaga',
    'version': '0.0.3.post1',
    'description': '',
    'long_description': None,
    'author': 'dokunogakusha',
    'author_email': '94387179+dokunogakusha@users.noreply.github.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



# Copyright (C) 2022 PlanQK
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from abc import ABC, abstractmethod
from typing import Generic

from dimod import SampleSet
from ..problem import ProblemType


class Terminator(Generic[ProblemType], ABC):
    """Interface for terminating an algorithm.

    """
    @abstractmethod
    def should_terminate(self, population: SampleSet) -> bool:
        """Checks termination condition for the current population.

        Returns:
            `True` to terminate, `False` to proceed.

        """
        pass


class Termination(Generic[ProblemType], ABC):
    """Interface to configure termination of an algorithm.
    """

    @abstractmethod
    def initialise(self, problem: ProblemType) -> Terminator[ProblemType]:
        """Problem-specific configuration of termination.

        Args:
            problem (Problem): A concrete problem instance, for which a
                bespoke termination condition can be prepared.

        Returns:
            A `Terminator[ProblemType]` implementation.

        """
        pass



# Copyright (C) 2022 PlanQK
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""This module contains all predefined classes to configure termination, i.e. if
a genetic algorithm should stop its iterative improvement of the population.

In general, one chooses one that guarantees termination, for instance after a
fix amount of iterations, and then combines it with (use case dependent) classes
that cover some early returns due to good population quality.

This module uses the abstract factory pattern to allow encapsulated state and
isolated optimisation runs. Implementations of `Termination` create
implementations of `Terminator`, that implement the actual functionality.

"""

from .termination import Terminator, Termination
from .generation_termination import GenerationTermination
from .and_termination import AndTerminator, AndTermination
from .or_termination import OrTerminator, OrTermination
from .lowest_energy_evaluation import LowestEnergyTerminator, LowestEnergyTermination



# Copyright (C) 2022 PlanQK
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""This module contains all predefined classes to configure selection, i.e. how
a genetic algorithm prepares a population to serve as input for the next generation.

Usually, it implements pruning to prevent large populations with bad individuals
and improve run time performance. In addition, one can use it to increase
diversity of population or even to integrate more constraints.

This module uses the abstract factory pattern to allow encapsulated state and
isolated optimisation runs. Implementations of `Selection` create
implementations of `Selector`, that implement the actual functionality.

"""

from .selection import Selector, Selection
from .identity_selection import IdentitySelector, IdentitySelection
from .truncation_selection import TruncationSelector, TruncationSelection
from .refilling_selection import RefillingSelector, RefillingSelection
from .pareto_selection import ParetoSelector, ParetoSelection



from abc import ABC, abstractmethod
from typing import Generic

from dimod import SampleSet
from ..problem import ProblemType


class Selector(Generic[ProblemType], ABC):
    """Interface for pruning a population.

    """
    @abstractmethod
    def select(self, population: SampleSet) -> SampleSet:
        """Prunes a population.

        Args:
            population (dimod.SampleSet): Nonempty population to be modified.

        Returns:
            A nonempty `dimod.SampleSet`.
        """
        pass


class Selection(Generic[ProblemType], ABC):
    """Interface to configure selection procedure of a genetic algorithm.

    It determines how the population of the previous iteration is adjusted
    before the next iteration.

    """
    @abstractmethod
    def initialise(self, problem: ProblemType) -> Selector[ProblemType]:
        """Problem-specific configuration of selection.

        Args:
            problem (ProblemType): A concrete problem instance,
                for which a bespoke selection can be prepared.

        Returns:
            A `Selector[ProblemType]` implementation.
        """
